from collections import OrderedDict
import json
import subprocess
from typing import Any, Collection, Dict, List, NewType, Sequence, Set, TypeVar

from typing_extensions import TypedDict

EnvVarName = NewType('EnvVarName', str)


class OpListItemsEntryOverview(TypedDict, total=False):
    tags: List[EnvVarName]


class OpListItemsEntry(TypedDict, total=False):
    overview: OpListItemsEntryOverview


OpListItemsOpaqueOutput = NewType('OpListItemsOpaqueOutput', List[Any])

FieldName = NewType('FieldName', str)

FieldValue = NewType('FieldValue', str)


class OPLookupError(LookupError):
    pass


class TooManyEntriesOPLookupError(OPLookupError):
    pass


class NoEntriesOPLookupError(OPLookupError):
    pass


class NoFieldValueOPLookupError(OPLookupError):
    pass


class InvalidTagOPLookupError(OPLookupError):
    pass


def _op_list_items(env_var_names: List[EnvVarName]) -> OpListItemsOpaqueOutput:
    list_command = ['op', 'list', 'items', '--tags',
                    ','.join(env_var_names)]
    list_items_json_docs_bytes = subprocess.check_output(list_command)
    # list_items_json_docs_str = list_items_json_docs_bytes.decode('utf-8')
    list_items_data: List[OpListItemsEntry] = json.loads(list_items_json_docs_bytes)
    by_env_var_name: Dict[EnvVarName, OpListItemsEntry] = {}
    for entry in list_items_data:
        for env_var_name in entry['overview']['tags']:
            if env_var_name in by_env_var_name:
                raise TooManyEntriesOPLookupError("Too many 1Password entries "
                                                  f"with tag {env_var_name} found")
            else:
                by_env_var_name[env_var_name] = entry
    ordered_list_items_data = []
    for env_var_name in env_var_names:
        if env_var_name not in by_env_var_name:
            raise NoEntriesOPLookupError(f"No 1Password entries with tag {env_var_name} found")
        else:
            ordered_list_items_data.append(by_env_var_name[env_var_name])
    return OpListItemsOpaqueOutput(ordered_list_items_data)


def _op_get_item(list_items_output: OpListItemsOpaqueOutput,
                 env_var_names: Collection[EnvVarName],
                 all_fields_to_seek: Collection[FieldName]) ->\
                 Dict[EnvVarName, Dict[FieldName, FieldValue]]:
    sorted_fields_to_seek = sorted(all_fields_to_seek)
    get_command: List[str] = ['op', 'get', 'item', '-', '--fields',
                              ','.join(sorted_fields_to_seek)]
    list_items_output_raw = json.dumps(list_items_output).encode('utf-8')
    field_values_json_docs_bytes = subprocess.check_output(get_command,
                                                           input=list_items_output_raw)
    field_values_json_docs_str = field_values_json_docs_bytes.decode('utf-8')
    field_values_data = [
        json.loads(field_values_json)
        for field_values_json
        in field_values_json_docs_str.split('\n')
        if field_values_json != ''
    ]
    return {
        env_var_name: field_values
        for (env_var_name, field_values)
        in zip(env_var_names, field_values_data)
    }


def _last_underscored_component_lowercased(env_var_name: EnvVarName) -> FieldName:
    components = env_var_name.split('_')
    return FieldName(components[-1].lower())


def _last_double_underscored_component_lowercased(env_var_name: EnvVarName) -> FieldName:
    components = env_var_name.split('__')
    return FieldName(components[-1].lower())


def _all_lowercased(env_var_name: EnvVarName) -> FieldName:
    return FieldName(env_var_name.lower())


T = TypeVar('T')


def _uniqify(fields: Sequence[T]) -> List[T]:
    "Removes duplicates but preserves order"
    # https://stackoverflow.com/questions/4459703/how-to-make-lists-contain-only-distinct-element-in-python
    return list(OrderedDict.fromkeys(fields))


def _aliases(fields: List[FieldName]) -> List[FieldName]:
    if 'user' in fields:
        return [FieldName('username')]
    else:
        return []


def _op_consolidated_fields(env_var_names: Collection[EnvVarName]) -> Set[FieldName]:
    return {
        field_name
        for env_var_name in env_var_names
        for field_name in _op_fields_to_try(env_var_name)
    }


def _op_fields_to_try(env_var_name: EnvVarName) -> List[FieldName]:
    candidates: List[FieldName] = _uniqify([
        _all_lowercased(env_var_name),
        _last_double_underscored_component_lowercased(env_var_name),
        _last_underscored_component_lowercased(env_var_name),
    ])
    return candidates + _aliases(candidates) + [FieldName('password')]


def _op_pluck_correct_field(env_var_name: EnvVarName,
                            field_values: Dict[FieldName,
                                               FieldValue]) -> FieldValue:
    fields = _op_fields_to_try(env_var_name)
    for field in fields:
        if field in field_values and field_values[field] != '':
            return field_values[field]
    raise NoFieldValueOPLookupError('1Passsword entry with tag '
                                    f'{env_var_name} has no value for '
                                    'the fields tried: '
                                    f'{", ".join(fields)}.  Please populate '
                                    'one of these fields in 1Password.')


def validate_env_var_names(env_var_names: List[EnvVarName]) -> None:
    for env_var_name in env_var_names:
        if ',' in env_var_name:
            raise InvalidTagOPLookupError('1Password does not support tags with commas')


def do_smart_lookups(env_var_names: List[EnvVarName]) -> Dict[str, str]:
    validate_env_var_names(env_var_names)
    list_items_output = _op_list_items(env_var_names)
    all_fields_to_seek = _op_consolidated_fields(env_var_names)
    field_values_for_envvars = _op_get_item(list_items_output,
                                            env_var_names,
                                            all_fields_to_seek)
    return {
        env_var_name: _op_pluck_correct_field(env_var_name, field_values_for_envvars[env_var_name])
        for env_var_name in field_values_for_envvars
    }



#!/usr/bin/env python

"""Tests for `op_env` package."""

import argparse
import io
import os
import subprocess
import tempfile
from typing import Dict
from unittest.mock import call, patch

import pytest
import yaml


from op_env._cli import Arguments, parse_argv, process_args
from op_env.op import (
    do_smart_lookups,
    _op_fields_to_try,
    _op_pluck_correct_field,
    EnvVarName,
    FieldName,
    FieldValue,
    InvalidTagOPLookupError,
    NoEntriesOPLookupError,
    NoFieldValueOPLookupError,
    TooManyEntriesOPLookupError,
)


@pytest.fixture
def list_of_number_yaml_file():
    with tempfile.NamedTemporaryFile(mode="w+t") as yaml_file:
        contents = [123, 456]
        yaml.dump(contents, yaml_file)
        yaml_file.flush()
        yield yaml_file.name


@pytest.fixture
def number_yaml_file():
    with tempfile.NamedTemporaryFile(mode="w+t") as yaml_file:
        contents = 123
        yaml.dump(contents, yaml_file)
        yaml_file.flush()
        yield yaml_file.name


@pytest.fixture
def string_yaml_file():
    with tempfile.NamedTemporaryFile(mode="w+t") as yaml_file:
        contents = 'foo'
        yaml.dump(contents, yaml_file)
        yaml_file.flush()
        yield yaml_file.name


@pytest.fixture
def object_yaml_file():
    with tempfile.NamedTemporaryFile(mode="w+t") as yaml_file:
        contents = {'foo': 'bar'}
        yaml.dump(contents, yaml_file)
        yaml_file.flush()
        yield yaml_file.name


@pytest.fixture
def invalid_yaml_file():
    with tempfile.NamedTemporaryFile(mode="w+t") as yaml_file:
        yaml_file.write('"')
        yaml_file.flush()
        yield yaml_file.name


@pytest.fixture
def empty_file():
    with tempfile.NamedTemporaryFile(mode="w+t") as yaml_file:
        yield yaml_file.name


@pytest.fixture
def one_item_yaml_file():
    with tempfile.NamedTemporaryFile(mode="w+t") as yaml_file:
        contents = ['VARA']
        yaml.dump(contents, yaml_file)
        yaml_file.flush()
        yield yaml_file.name


@pytest.fixture
def two_item_yaml_file():
    with tempfile.NamedTemporaryFile(mode="w+t") as yaml_file:
        contents = ['VAR1', 'VAR2']
        yaml.dump(contents, yaml_file)
        yaml_file.flush()
        yield yaml_file.name


def test_process_args_shows_json_with_simple_env() -> None:
    with patch('op_env.op._op_list_items') as mock_op_list_items,\
         patch('op_env.op._op_consolidated_fields') as mock_op_consolidated_fields,\
         patch('op_env.op._op_get_item') as mock_op_get_item,\
         patch('op_env.op._op_pluck_correct_field') as mock_op_pluck_correct_field,\
         patch('sys.stdout', new_callable=io.StringIO) as stdout_stringio:
        mock_list_items_output = mock_op_list_items.return_value
        mock_all_fields_to_seek = mock_op_consolidated_fields.return_value
        Dict[EnvVarName, Dict[FieldName, FieldValue]]
        retval: Dict[EnvVarName, Dict[FieldName, FieldValue]] = {
            EnvVarName('a'): {
                FieldName('password'): FieldValue('1'),
            }}
        mock_op_get_item.return_value = retval
        env_var_names = [EnvVarName('a')]
        args: Arguments = {
            'operation': 'json',
            'environment': env_var_names,
            'command': []
        }
        mock_op_pluck_correct_field.return_value = '1'
        process_args(args)
        assert stdout_stringio.getvalue() == '{"a": "1"}\n'
        mock_op_list_items.assert_called_with(env_var_names)
        mock_op_consolidated_fields.assert_called_with(env_var_names)
        mock_op_pluck_correct_field.assert_called_with('a', {'password': '1'})
        mock_op_get_item.assert_called_with(mock_list_items_output,
                                            env_var_names,
                                            mock_all_fields_to_seek)


def test_process_args_runs_simple_command_with_simple_env():
    with patch('op_env._cli.subprocess') as mock_subprocess,\
         patch('op_env._cli.do_smart_lookups') as mock_do_smart_lookups,\
         patch.dict(os.environ, {'ORIGINAL_ENV': 'TRUE'}, clear=True):
        command = ['env']
        args = {'operation': 'run', 'command': command,
                'environment': ['a']}
        mock_do_smart_lookups.return_value = {'a': '1'}
        process_args(args)
        mock_do_smart_lookups.assert_called_with(['a'])
        mock_subprocess.check_call.assert_called_with(command,
                                                      env={'a': '1',
                                                           'ORIGINAL_ENV': 'TRUE'})


def test_process_args_shows_env_with_variables_needing_escape():
    with patch('op_env._cli.do_smart_lookups') as mock_do_smart_lookups,\
         patch.dict(os.environ, {'ORIGINAL_ENV': 'TRUE'}, clear=True),\
         patch('sys.stdout', new_callable=io.StringIO) as stdout_stringio:
        args = {'operation': 'sh', 'environment': ['a', 'c']}
        mock_do_smart_lookups.return_value = {'a': "'", 'c': 'd'}
        process_args(args)
        assert stdout_stringio.getvalue() == 'a=\'\'"\'"\'\'; export a\nc=d; export c\n'


def test_process_args_shows_env_with_multiple_variables():
    def fake_op_smart_lookup(k):
        return {
            'a': 'b',
            'c': 'd',
         }[k]

    with patch('op_env._cli.do_smart_lookups') as mock_do_smart_lookups,\
         patch.dict(os.environ, {'ORIGINAL_ENV': 'TRUE'}, clear=True),\
         patch('sys.stdout', new_callable=io.StringIO) as stdout_stringio:
        mock_do_smart_lookups.return_value = {'a': 'b', 'c': 'd'}
        args = {'operation': 'sh', 'environment': ['a', 'c']}
        process_args(args)
        assert stdout_stringio.getvalue() == 'a=b; export a\nc=d; export c\n'


def test_process_args_shows_env_with_simple_env():
    with patch('op_env._cli.do_smart_lookups') as mock_do_smart_lookups,\
         patch.dict(os.environ, {'ORIGINAL_ENV': 'TRUE'}, clear=True),\
         patch('sys.stdout', new_callable=io.StringIO) as stdout_stringio:
        mock_do_smart_lookups.return_value = {'a': 'b'}
        args = {'operation': 'sh', 'environment': ['a']}
        process_args(args)
        assert stdout_stringio.getvalue() == 'a=b; export a\n'


@pytest.mark.skip(reason="need to mock op binary in test PATH")
def test_process_args_runs_simple_command():
    with patch('op_env._cli.subprocess') as mock_subprocess,\
         patch.dict(os.environ, {'ORIGINAL_ENV': 'TRUE'}, clear=True):
        command = ['env']
        args = {'operation': 'run', 'command': command,
                'environment': []}
        process_args(args)
        mock_subprocess.check_call.assert_called_with(command, env={
            'ORIGINAL_ENV': 'TRUE',
        })


def test_process_args_rejects_non_run():
    with patch('op_env.op.subprocess'):  # for safety
        with pytest.raises(ValueError):
            args = {'operation': 'definitely-not-run'}
            process_args(args)


def test_fields_to_try_breaks_on_double_underscore_and_underscore():
    with patch('op_env.op.subprocess'):  # for safety
        out = _op_fields_to_try('ABC__FLOOGLE_BAR')
        assert out == ['abc__floogle_bar', 'floogle_bar', 'bar', 'password']


def test_fields_to_try_breaks_on_double_underscore():
    with patch('op_env.op.subprocess'):  # for safety
        out = _op_fields_to_try('ABC__FLOOGLE')
        assert out == ['abc__floogle', 'floogle', 'password']


def test_fields_to_try_conversion_username():
    with patch('op_env.op.subprocess'):  # for safety
        out = _op_fields_to_try('ABC_USER')
        assert out == ['abc_user', 'user', 'username', 'password']


def test_fields_to_try_multiple_words():
    with patch('op_env.op.subprocess'):  # for safety
        out = _op_fields_to_try('ABC_FLOOGLE')
        assert out == ['abc_floogle', 'floogle', 'password']


def test_fields_to_try_simple():
    with patch('op_env.op.subprocess'):  # for safety
        out = _op_fields_to_try('ABC')
        assert out == ['abc', 'password']


def test_op_do_smart_lookups_multiple_entries():
    with patch('op_env.op.subprocess') as mock_subprocess:
        list_output = \
            b'[{"overview":{"tags":["ANY_TEST_VALUE"]}},' \
            b'{"overview": {"tags": ["ANOTHER_TEST_VALUE"]}}]'
        post_processed_list_output = \
            b'[{"overview": {"tags": ["ANY_TEST_VALUE"]}}, ' \
            b'{"overview": {"tags": ["ANOTHER_TEST_VALUE"]}}]'
        get_output = \
            b'{"any_test_value":"","another_test_value":"","password":"any","value":""}\n' \
            b'{"any_test_value":"","another_test_value":"another",'\
            b'"password":"get_results","value":""}'
        mock_subprocess.check_output.side_effect = [
            list_output,
            get_output,
        ]
        out = do_smart_lookups(['ANY_TEST_VALUE', 'ANOTHER_TEST_VALUE'])
        mock_subprocess.check_output.\
            assert_has_calls([call(['op', 'list', 'items', '--tags',
                                    'ANY_TEST_VALUE,ANOTHER_TEST_VALUE']),
                              call(['op', 'get', 'item', '-', '--fields',
                                    'another_test_value,any_test_value,password,value'],
                                   input=post_processed_list_output)])
        assert out == {
            'ANY_TEST_VALUE': 'any',
            'ANOTHER_TEST_VALUE': 'another'
        }


def test_do_smart_lookups_no_field_value():
    with patch('op_env.op.subprocess') as mock_subprocess:
        list_output = b'[{"overview":{"tags":["ANY_TEST_VALUE"]}}]'
        post_processed_list_output = b'[{"overview": {"tags": ["ANY_TEST_VALUE"]}}]'
        get_output = b'{"password":""}\n'
        mock_subprocess.check_output.side_effect = [
            list_output,
            get_output,
        ]
        with pytest.raises(NoFieldValueOPLookupError,
                           match=('1Passsword entry with tag ANY_TEST_VALUE '
                                  'has no value for the fields tried: '
                                  'any_test_value, value, password.  '
                                  'Please populate one of these fields in 1Password.')):
            do_smart_lookups(['ANY_TEST_VALUE'])
        mock_subprocess.check_output.\
            assert_has_calls([call(['op', 'list', 'items', '--tags', 'ANY_TEST_VALUE']),
                              call(['op', 'get', 'item', '-', '--fields',
                                    'any_test_value,password,value'],
                                   input=post_processed_list_output)])


def test_do_smart_lookups_too_few_entries():
    with patch('op_env.op.subprocess') as mock_subprocess:
        list_output = b"[]"
        mock_subprocess.check_output.return_value = list_output
        with pytest.raises(NoEntriesOPLookupError,
                           match='No 1Password entries with tag ANY_TEST_VALUE found'):
            do_smart_lookups(['ANY_TEST_VALUE'])
        mock_subprocess.check_output.\
            assert_called_with(['op', 'list', 'items', '--tags', 'ANY_TEST_VALUE'])


def test_do_smart_lookups_too_many_entries():
    with patch('op_env.op.subprocess') as mock_subprocess:
        list_output = \
            b'[{"overview":{"tags":["ANY_TEST_VALUE"]}},{"overview":{"tags":["ANY_TEST_VALUE"]}}]'
        mock_subprocess.check_output.return_value = list_output
        with pytest.raises(TooManyEntriesOPLookupError,
                           match='Too many 1Password entries with tag ANY_TEST_VALUE'):
            do_smart_lookups(['ANY_TEST_VALUE'])
        mock_subprocess.check_output.\
            assert_called_with(['op', 'list', 'items', '--tags', 'ANY_TEST_VALUE'])


def test_op_do_smart_lookups_comma_in_env():
    with patch('op_env.op.subprocess') as mock_subprocess:
        list_output = b'[{"overview": {"tags": ["ANY_TEST_VALUE"]}}]'
        get_output = b'{"any_test_value":"","password":"get_results","value":""}\n'
        mock_subprocess.check_output.side_effect = [
            list_output,
            get_output,
        ]
        with pytest.raises(InvalidTagOPLookupError,
                           match='1Password does not support tags with commas'):
            do_smart_lookups(['ENV_WITH_,_IN_IT'])
        mock_subprocess.check_output.assert_not_called()


def test_op_do_smart_lookups_one_var():
    with patch('op_env.op.subprocess') as mock_subprocess:
        list_output = b'[{"overview": {"tags": ["ANY_TEST_VALUE"]}}]'
        get_output = b'{"any_test_value":"","password":"get_results","value":""}\n'
        mock_subprocess.check_output.side_effect = [
            list_output,
            get_output,
        ]
        out = do_smart_lookups(['ANY_TEST_VALUE'])
        mock_subprocess.check_output.\
            assert_has_calls([call(['op', 'list', 'items', '--tags', 'ANY_TEST_VALUE']),
                              call(['op', 'get', 'item', '-', '--fields',
                                    'any_test_value,password,value'],
                                   input=list_output)])
        assert out == {'ANY_TEST_VALUE': 'get_results'}


def test_op_pluck_correct_field_multiple_fields():
    with patch('op_env.op._op_fields_to_try') as mock_op_fields_to_try:
        mock_op_fields_to_try.return_value = ['floogle', 'blah']
        ret = _op_pluck_correct_field('ENVVARNAME', {'blah': '', 'floogle': 'result value'})
        mock_op_fields_to_try.assert_called_with('ENVVARNAME')
        assert ret == 'result value'


def test_op_pluck_correct_field_multiple_fields_all_errors():
    with patch('op_env.op._op_fields_to_try') as mock_op_fields_to_try:
        mock_op_fields_to_try.return_value = ['floogle', 'blah']
        with pytest.raises(NoFieldValueOPLookupError,
                           match=('1Passsword entry with tag '
                                  'ENVVARNAME has no value for '
                                  'the fields tried: '
                                  "floogle, blah.  Please populate "
                                  'one of these fields in 1Password.')):
            _op_pluck_correct_field('ENVVARNAME', {'floogle': '', 'blah': ''})
        mock_op_fields_to_try.assert_called_with('ENVVARNAME')


def test_op_pluck_correct_field_single_field_with_error():
    with patch('op_env.op._op_fields_to_try') as mock_op_fields_to_try:
        mock_op_fields_to_try.return_value = ['floogle']
        with pytest.raises(NoFieldValueOPLookupError):
            _op_pluck_correct_field('ENVVARNAME', {'floogle': ''})
        mock_op_fields_to_try.assert_called_with('ENVVARNAME')


def test_op_pluck_correct_field_multiple_fields_chooses_second():
    with patch('op_env.op._op_fields_to_try') as mock_op_fields_to_try:
        mock_op_fields_to_try.return_value = ['floogle', 'blah']
        ret = _op_pluck_correct_field('ENVVARNAME', {'floogle': '', 'blah': 'result value'})
        mock_op_fields_to_try.assert_called_with('ENVVARNAME')
        assert ret == 'result value'


def test_op_pluck_correct_field_chooses_first():
    with patch('op_env.op._op_fields_to_try') as mock_op_fields_to_try:
        mock_op_fields_to_try.return_value = ['floogle']
        ret = _op_pluck_correct_field('ENVVARNAME', {'floogle': 'myvalue'})
        mock_op_fields_to_try.assert_called_with('ENVVARNAME')
        assert ret == 'myvalue'


def test_parse_args_json_operation_no_env_variables():
    argv = ['op-env', 'json']
    args = parse_argv(argv)
    assert args == {'environment': [],
                    'operation': 'json'}


def test_parse_args_run_operation_with_long_env_variables():
    argv = ['op-env', 'run', '-e', 'DUMMY', '--environment', 'DUMMY2', 'mycmd']
    args = parse_argv(argv)
    assert args == {'command': ['mycmd'],
                    'environment': ['DUMMY', 'DUMMY2'],
                    'operation': 'run'}


def test_parse_args_run_operation_no_env_variables():
    argv = ['op-env', 'run', 'mycmd']
    args = parse_argv(argv)
    assert args == {'command': ['mycmd'],
                    'environment': [],
                    'operation': 'run'}


def test_parse_args_run_operation_with_multiple_environment_arguments():
    argv = ['op-env', 'run', '-e', 'DUMMY', '-e', 'DUMMY2', 'mycmd']
    args = parse_argv(argv)
    assert args == {'command': ['mycmd'],
                    'environment': ['DUMMY', 'DUMMY2'],
                    'operation': 'run'}


def test_parse_args_run_operation_with_environment_arguments():
    argv = ['op-env', 'run', '-e', 'DUMMY', 'mycmd', '1', '2', '3']
    args = parse_argv(argv)
    assert args == {'command': ['mycmd', '1', '2', '3'],
                    'environment': ['DUMMY'],
                    'operation': 'run'}


def test_parse_args_run_operation_with_multiple_yaml_and_environment_arguments(one_item_yaml_file,
                                                                               two_item_yaml_file):
    argv = ['op-env', 'run', '-e', 'VAR_1', '-e', 'VAR0',
            '-y', two_item_yaml_file, '-y', one_item_yaml_file,
            'mycmd', '1', '2', '3']
    args = parse_argv(argv)
    assert args == {'command': ['mycmd', '1', '2', '3'],
                    'environment': ['VAR_1', 'VAR0', 'VAR1', 'VAR2', 'VARA'],
                    'operation': 'run'}


def test_parse_args_run_operation_with_yaml_arguments_and_environment_arguments(two_item_yaml_file):
    argv = ['op-env', 'run', '-e', 'VAR0', '-y', two_item_yaml_file, 'mycmd', '1', '2', '3']
    args = parse_argv(argv)
    assert args == {'command': ['mycmd', '1', '2', '3'],
                    'environment': ['VAR0', 'VAR1', 'VAR2'],
                    'operation': 'run'}


def test_list_of_numbers_yaml_argument(list_of_number_yaml_file):
    argv = ['op-env', 'run', '-y', list_of_number_yaml_file, 'mycmd', '1', '2', '3']
    with pytest.raises(argparse.ArgumentTypeError,
                       match='YAML file must contain a list of strings'):
        parse_argv(argv)


def test_parse_args_run_operation_with_number_file_yaml_argument(number_yaml_file):
    argv = ['op-env', 'run', '-y', number_yaml_file, 'mycmd', '1', '2', '3']
    with pytest.raises(argparse.ArgumentTypeError, match='YAML file must be a list; found'):
        parse_argv(argv)


def test_parse_args_run_operation_with_string_file_yaml_argument(string_yaml_file):
    argv = ['op-env', 'run', '-y', string_yaml_file, 'mycmd', '1', '2', '3']
    with pytest.raises(argparse.ArgumentTypeError, match='YAML file must be a list; found'):
        parse_argv(argv)


def test_parse_args_run_operation_with_object_file_yaml_argument(object_yaml_file):
    argv = ['op-env', 'run', '-y', object_yaml_file, 'mycmd', '1', '2', '3']
    with pytest.raises(argparse.ArgumentTypeError, match='YAML file must be a list; found'):
        parse_argv(argv)


def test_parse_args_run_operation_with_invalid_file_yaml_argument(invalid_yaml_file):
    argv = ['op-env', 'run', '-y', invalid_yaml_file, 'mycmd', '1', '2', '3']
    with pytest.raises(yaml.scanner.ScannerError):
        parse_argv(argv)


def test_parse_args_run_operation_with_empty_file_yaml_argument(empty_file):
    argv = ['op-env', 'run', '-y', empty_file, 'mycmd', '1', '2', '3']
    args = parse_argv(argv)
    assert args == {'command': ['mycmd', '1', '2', '3'],
                    'environment': [],
                    'operation': 'run'}


def test_parse_args_run_operation_with_yaml_argument(two_item_yaml_file):
    argv = ['op-env', 'run', '-y', two_item_yaml_file, 'mycmd', '1', '2', '3']
    args = parse_argv(argv)
    assert args == {'command': ['mycmd', '1', '2', '3'],
                    'environment': ['VAR1', 'VAR2'],
                    'operation': 'run'}


def test_parse_args_run_simple():
    argv = ['op-env', 'run', '-e', 'DUMMY', 'mycmd']
    args = parse_argv(argv)
    assert args == {'command': ['mycmd'], 'environment': ['DUMMY'], 'operation': 'run'}


def test_parse_args_sh_simple():
    argv = ['op-env', 'sh', '-e', 'DUMMY']
    args = parse_argv(argv)
    assert args == {'environment': ['DUMMY'], 'operation': 'sh'}


@pytest.mark.skip(reason="need to mock op binary in test PATH")
def test_cli_run():
    argv = ['op-env', 'run', '-e', 'DUMMY', 'env']
    expected_envvar = 'DUMMY=dummyvalue'
    actual_output = subprocess.check_output(argv).decode('utf-8')
    assert expected_envvar in actual_output


def test_cli_help_run():
    expected_help = """usage: op-env run [-h] [--environment ENVVAR] [--yaml-environment YAMLENV] command [command ...]

Run the specified command with the given environment variables

positional arguments:
  command               Command to run with the environment set from 1Password

optional arguments:
  -h, --help            show this help message and exit
  --environment ENVVAR, -e ENVVAR
                        environment variable name to set, based on item with same tag in 1Password
  --yaml-environment YAMLENV, -y YAMLENV
                        YAML config specifying a list of environment variable names to set
"""
    request_long_lines = {'COLUMNS': '999', 'LINES': '25'}
    env = {}
    env.update(os.environ)
    env.update(request_long_lines)

    # older python versions show arguments like this:
    actual_help = subprocess.check_output(['op-env', 'run', '--help'], env=env).decode('utf-8')
    assert actual_help == expected_help


def test_cli_help_json():
    expected_help = """usage: op-env json [-h] [--environment ENVVAR] [--yaml-environment YAMLENV]

Produce simple JSON on stdout mapping requested env variables to values

optional arguments:
  -h, --help            show this help message and exit
  --environment ENVVAR, -e ENVVAR
                        environment variable name to set, based on item with same tag in 1Password
  --yaml-environment YAMLENV, -y YAMLENV
                        YAML config specifying a list of environment variable names to set
"""
    request_long_lines = {'COLUMNS': '999', 'LINES': '25'}
    env = {}
    env.update(os.environ)
    env.update(request_long_lines)

    # older python versions show arguments like this:
    actual_help = subprocess.check_output(['op-env', 'json', '--help'], env=env).decode('utf-8')
    assert actual_help == expected_help


def test_cli_help_sh():
    expected_help = """usage: op-env sh [-h] [--environment ENVVAR] [--yaml-environment YAMLENV]

Produce commands on stdout that can be 'eval'ed to set variables in current shell

optional arguments:
  -h, --help            show this help message and exit
  --environment ENVVAR, -e ENVVAR
                        environment variable name to set, based on item with same tag in 1Password
  --yaml-environment YAMLENV, -y YAMLENV
                        YAML config specifying a list of environment variable names to set
"""
    request_long_lines = {'COLUMNS': '999', 'LINES': '25'}
    env = {}
    env.update(os.environ)
    env.update(request_long_lines)

    # older python versions show arguments like this:
    actual_help = subprocess.check_output(['op-env', 'sh', '--help'], env=env).decode('utf-8')
    assert actual_help == expected_help


def test_cli_no_args():
    expected_help = """usage: op-env [-h] {run,json,sh} ...
op-env: error: the following arguments are required: operation
"""
    request_long_lines = {'COLUMNS': '999', 'LINES': '25'}
    env = {}
    env.update(os.environ)
    env.update(request_long_lines)

    # older python versions show arguments like this:
    completed_process = subprocess.run(['op-env'], env=env,
                                       stderr=subprocess.PIPE)
    actual_help = completed_process.stderr.decode('utf-8')
    assert actual_help == expected_help
    assert completed_process.returncode == 2


def test_cli_help():
    expected_help = """usage: op-env [-h] {run,json,sh} ...

positional arguments:
  {run,json,sh}
    run          Run the specified command with the given environment variables
    json         Produce simple JSON on stdout mapping requested env variables to values
    sh           Produce commands on stdout that can be 'eval'ed to set variables in current shell

optional arguments:
  -h, --help     show this help message and exit
"""
    request_long_lines = {'COLUMNS': '999', 'LINES': '25'}
    env = {}
    env.update(os.environ)
    env.update(request_long_lines)

    # older python versions show arguments like this:
    actual_help = subprocess.check_output(['op-env', '--help'],
                                          env=env).decode('utf-8')
    assert actual_help == expected_help



"""Console script for op_env."""
import argparse
import json
import os
import pipes
import subprocess
import sys
from typing import Any, List, Optional, Sequence, Union

from typing_extensions import TypedDict
import yaml

from .op import do_smart_lookups, EnvVarName


class Arguments(TypedDict):
    operation: str
    environment: List[EnvVarName]
    command: List[str]


class AppendListFromYAMLAction(argparse.Action):
    def __call__(self,
                 parser: argparse.ArgumentParser,
                 namespace: argparse.Namespace,
                 values: Union[str, Sequence[Any], None],
                 option_string: Optional[str] = None):
        assert isinstance(values, str)  # should be validated already by argparse
        filename = values
        with open(filename, 'r') as stream:
            variables_from_yaml = yaml.safe_load(stream)
            if variables_from_yaml is None:
                # treat an empty file as an empty list
                variables_from_yaml = []
            if not isinstance(variables_from_yaml, list):
                raise argparse.ArgumentTypeError('YAML file must be a list; '
                                                 f'found {variables_from_yaml}')
            if not all([isinstance(item, str) for item in variables_from_yaml]):
                raise argparse.ArgumentTypeError('YAML file must contain a list of strings; '
                                                 f'found {variables_from_yaml}')
        envvars = getattr(namespace, self.dest)
        envvars.extend(variables_from_yaml)


def add_environment_arguments(arg_parser: argparse.ArgumentParser):
    arg_parser.add_argument('--environment', '-e',
                            metavar='ENVVAR',
                            action='append',
                            default=[],
                            help='environment variable name to set, '
                            'based on item with same tag in 1Password')
    arg_parser.add_argument('--yaml-environment', '-y',
                            metavar='YAMLENV',
                            action=AppendListFromYAMLAction,
                            dest='environment',
                            default=[],
                            help='YAML config specifying a list of environment variable '
                            'names to set')


def parse_argv(argv: List[str]) -> Arguments:
    parser = argparse.ArgumentParser()

    subparsers = parser.add_subparsers(dest='operation')
    subparsers.required = True
    run_desc = 'Run the specified command with the given environment variables'
    run_parser = subparsers.add_parser('run',
                                       description=run_desc,
                                       help=run_desc)
    add_environment_arguments(run_parser)
    run_parser.add_argument('command',
                            nargs='+',
                            help='Command to run with the environment set from 1Password')
    json_desc = 'Produce simple JSON on stdout mapping requested env variables to values'
    json_parser = subparsers.add_parser('json',
                                        description=json_desc,
                                        help=json_desc)
    add_environment_arguments(json_parser)
    sh_desc = ("Produce commands on stdout that can be 'eval'ed to set "
               "variables in current shell")
    sh_parser = subparsers.add_parser('sh',
                                      help=sh_desc,
                                      description=sh_desc)
    add_environment_arguments(sh_parser)
    return vars(parser.parse_args(argv[1:]))  # type: ignore


def process_args(args: Arguments) -> int:
    if args['operation'] == 'run':
        copied_env = dict(os.environ)
        new_env = do_smart_lookups(args['environment'])
        copied_env.update(new_env)
        subprocess.check_call(args['command'], env=copied_env)
        return 0
    elif args['operation'] == 'json':
        new_env = do_smart_lookups(args['environment'])
        print(json.dumps(new_env))
        return 0
    elif args['operation'] == 'sh':
        new_env = do_smart_lookups(args['environment'])
        for envvar, envvalue in new_env.items():
            print(f'{envvar}={pipes.quote(envvalue)}; export {envvar}')
        return 0
    else:
        raise ValueError(f"Unknown operation: {args['operation']}")


def main(argv: List[str] = sys.argv) -> int:
    """Console script for op_env."""
    args = parse_argv(argv)
    return process_args(args)


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""The setup script."""
from setuptools import setup, find_packages
from typing import List

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements: List[str] = ['typing_extensions', 'PyYAML']

setup_requirements: List[str] = ['pytest-runner']

test_requirements: List[str] = ['pytest>=3']


setup(
    author="Vince Broz",
    author_email='vince@broz.cc',
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    description="op-env allows you to use 1Password entries as environment variable-style secrets",  # noqa: E501
    entry_points={
        'console_scripts': [
            'op-env=op_env._cli:main',
        ],
    },
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/x-rst',
    include_package_data=True,
    keywords='op_env',
    name='op_env',
    packages=find_packages(include=['op_env',
                                    'op_env.*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/apiology/op_env',
    version='0.6.0',
    zip_safe=False,
)
